// -*- Mode:C++ -*-
//Header:
//File: fsringT.h
//Author: NODA, Itsuki
//Date: 1999/05/18
//

//ModifyHistory:
// 1999/05/18: Start to create this file
//EndModifyHistory:

/*
 * Copyright (C) 2001 NODA, Itsuki, CARC, AIST, JAPAN
 * Copyright (C) 1999, 2000 Itsuki Noda, Electrotechnical Laboratory, Japan
 */

#ifndef _itk_RingT_h_
#define _itk_RingT_h_
////////////////////////////////////////////////////////////////////////

#include "itk/btype.h"
#include "itk/RcylHeapT.h"

namespace Itk {

//======================================================================
// Ring<Content> : cyclic double linked list
//

   template <class Content>
      class RingT : public WithDescriber {
	 public:
	    //--------------------------------------------------
	    // local typedef

	    typedef RingT<Content> Ring ;
 
	    //--------------------------------------------------
	    // member

	    Content entity ;
	    Ring *prev ;
	    Ring *next ;

	    //--------------------------------------------------
	    // constructor

	    RingT<Content>() {
	       prev = ITK_NULLPTR ;
	       next = ITK_NULLPTR ;
	    } ;
	    RingT<Content>(const Content& val, Bool circlep = True) {
	       entity = val ;
	       if(circlep) {
		  linkSelf() ;
	       } else {
		  prev = ITK_NULLPTR ;
		  next = ITK_NULLPTR ;
	       }
	    } ;
	    RingT<Content>(const RingT<Content> & org) {
	       copy(org) ;
	    } ;

	    //--------------------------------------------------
	    // copy
      
	    Ring & copy(const Ring & org) {
	       entity = org.entity ;
	       prev = org.prev ;
	       next = org.next ;
	       return *this ;
	    } ;

	    //--------------------------------------------------
	    // set value

	    Ring * set(const Content & val) {
	       entity = val ;
	       return this ;
	    } ;

	    //--------------------------------------------------
	    // link next

	    Ring * linkNext(Ring* newnext) {
	       next = newnext ;
	       newnext->prev = this ;
	       return this ;
	    } ;

	    //--------------------------------------------------
	    // link prev

	    Ring * linkPrev(Ring* newprev) {
	       return newprev->linkNext(this) ;
	    } ;

	    //--------------------------------------------------
	    // link self (single ring)

	    Ring * linkSelf() { return linkNext(this) ; } ;

	    //--------------------------------------------------
	    // make single ring

	    Ring * setSingle(const Content & val) {
	       set(val) ;
	       return linkSelf() ;
	    } ;

	    //--------------------------------------------------
	    // single ring check

	    Bool isSingle() { return next == this && prev == this ; } ;

	    //--------------------------------------------------
	    // insert next

	    Ring* insertNext(Ring* newnext) {
	       Ring * oldnext = next ;
	       Ring * oldnewnextprev = newnext->prev ;

	       linkNext(newnext) ;
	       oldnext->linkPrev(oldnewnextprev) ;
	       return this ;
	    } ;

	    Ring* insertFirst(Ring* ring) { return insertNext(ring) ; } ;

	    //--------------------------------------------------
	    // insert prev

	    Ring* insertPrev(Ring* newprev) {
	       Ring * oldprev = prev ;
	       Ring * oldnewprevnext = newprev->next ;
	
	       linkPrev(newprev) ;
	       oldprev->linkNext(oldnewprevnext) ;

	       return this ;
	    } ;

	    Ring* insertLast(Ring* ring) { return insertPrev(ring) ; } ;

	    //--------------------------------------------------
	    // add insert new data (allocate new memory)

	    Ring* putFirst(const Content& dt) {
	       Ring * newring = new Ring(dt,True) ;
	       insertNext(newring) ;
	       return newring ;
	    } ;

	    Ring* putLast(const Content& dt) {
	       Ring* newring = new Ring(dt,True) ;
	       insertPrev(newring) ;
	       return newring ;
	    } ;

	    //--------------------------------------------------
	    // remove self 

	    void removeSelf() {
	       Ring *pre = prev ;
	       Ring *nxt = next ;
	       pre->next = nxt ;
	       nxt->prev = pre ;
	    } ;

	    //--------------------------------------------------
	    // remove one
      
	    void remove(Ring* del) { del->removeSelf() ; } ;

	    //--------------------------------------------------
	    // split (split operation is same as insert operation)

	    Ring * splitNext(Ring * newnext) {
	       return insertNext(newnext) ;
	    } ;
	 
	    Ring * splitPrev(Ring * newprev) {
	       return insertPrev(newprev) ;
	    } ;

	    //--------------------------------------------------
	    // end check

	    Bool isEnd(const Ring* tail) const { return tail == this ; } ;

	    //--------------------------------------------------
	    // reverse

	    void reverse() ;

	    //--------------------------------------------------
	    // length

	    UInt length() const ;

	    //--------------------------------------------------
	    // describe

	    virtual void describe(ostream& ostr, 
				  const Bool detailp = True) const {
	       ostr << "#Ring[<" << this << ">:<" << prev << ">=*=<" 
		    << next << ">]" ;
	       if(detailp) {
		  const Ring *r = this ;
		  ostr << endl << "  [Content]" ;
		  do {
		     ostr << endl << "\t" << *r << ": " << r->entity ;
		     r = r->next ;
		  } while (!isEnd(r)) ;
		  ostr << endl << "\t--- end ---" << endl ;
	       }
	    } ;
      } ;      
   
//--------------------------------------------------
// reverse

   template <class Content>
      void RingT<Content>::reverse() {
      Ring r = this ;
      Ring nr ;
      do {
	 nr = r->next ;
	 r->next= r->prev ;
	 r->prev = nr ;
	 r = nr ;
      } while (!isEnd(r)) ;
      return this ;
   }

//--------------------------------------------------
// length

   template <class Content>
      UInt RingT<Content>::length() const {
      const Ring* r = this ;
      UInt l = 0 ;
      do {
	 r = r->next ;
	 l++ ;
      } while (!isEnd(r)) ;
      return l ;
   }

//======================================================================
// RingRcylHeap<Content> : ring heap
//

#ifndef ITK_DFLT_RINGRCYLHEAP_NAME
#define ITK_DFLT_RINGRCYLHEAP_NAME		"RingRcylHeap"
#endif


   template <class Content>
      class RingRcylHeapT : public RcylHeapT< RingT<Content> > {
	 public:
	    //--------------------------------------------------
	    // typedef 

	    typedef RingT<Content> Ring ;
	    typedef RingRcylHeapT<Content> RingRcylHeap ;
      
	    //--------------------------------------------------
	    // constructor
      
	    RingRcylHeapT<Content>(char *nm = ITK_DFLT_RINGRCYLHEAP_NAME,
				   UInt sz = ITK_DFLT_MEMARRAY_SIZE,
				   UInt ef = ITK_DFLT_EXPAND_FACTOR) {
	       init(nm,sz,ef) ;
	    } ;

	    //--------------------------------------------------
	    // get

	    Ring * getPtr() {
	       Ring * r = RcylHeap::getPtr() ;
	       r->linkSelf() ;
	       return r ;
	    } ;
      
	    Ring & get() { return *getPtr() ; } ;

	    //--------------------------------------------------
	    // get & set
      
	    Ring * getPtr(const Content& val) {
	       Ring * r = getPtr() ;
	       return r->set(val) ;
	    } ;

	    Ring & get(const Content& val) { return *getPtr(val) ; } ;

	    //--------------------------------------------------
	    // add ring at first position

	    Ring * addRingFirst(Ring* root) {
	       Ring * r = getPtr() ;
	       root->insertNext(r) ;
	       return r ;
	    } ;
      
	    Ring * addRingFirst(Ring* root, const Content& val) {
	       Ring * r = getPtr(val) ;
	       root->insertNext(r) ;
	       return r ;
	    } ;

	    //--------------------------------------------------
	    // add ring at last position

	    Ring * addRingLast(Ring* root) {
	       Ring * r = getPtr() ;
	       root->insertPrev(r) ;
	       return r ;
	    } ;
      
	    Ring * addRingLast(Ring* root, const Content& val) {
	       Ring * r = getPtr(val) ;
	       root->insertPrev(r) ;
	       return r ;
	    } ;
      
      } ;

} ;

////////////////////////////////////////////////////////////////////////
#endif


	 

      
      
      

